# -*- coding: utf-8 -*-

from functools import wraps
from threading import Thread
import socket

def singleton(o_cls):
    orig_new = o_cls.__new__
    inst_field = "__instance"

    @wraps(o_cls.__new__)
    def __new__(cls, *args, **kwargs):
        if (instance := getattr(cls, inst_field, None)) is None:
            instance = orig_new(cls)
            if hasattr(cls, "__init__"):
                cls.__init__(instance, *args, **kwargs)
                delattr(cls, "__init__")
            setattr(cls, inst_field, instance)
        return instance

    o_cls.__new__ = __new__
    return o_cls


@singleton
class Delivery:
    ''' Un "postino" tra moduli'''
    
    def __init__(self):
        self.subscr = {}
    
    def subscribe(self, group: str, method: any) -> None:
        ''' Aggiunge un subscriber ad un dato gruppo '''
        if not group in self.subscr.keys():
            self.subscr[group] = []
        if method in self.subscr[group]:
            raise RuntimeError('Metodo già acquisito nel gruppo %s' % group)
        self.subscr[group].append(method)

    def unsubscribe(self, group: str, method: any) -> None:
        ''' Rimuove un subscriber ad un dato gruppo, i gruppi vuoti vengono rimossi'''
        if not group in self.subscr.keys():
            raise RuntimeError('Gruppo %s non registrato' % group)
        if method not in self.subscr[group]:
            raise RuntimeError('Metodo non registrato nel gruppo %s' % group)
        self.subscr[group].remove(method)
        if not self.subscr[group]:
            del self.subscr[group]
    
    def send_message(self, group:str, message: list) -> None:
        ''' Distribuisce un messaggio ai subscriber di un gruppo '''
        if not group in self.subscr.keys():
            raise RuntimeError('Gruppo %s non registrato' % group)
        for s_func in self.subscr[group]:
            s_func(message)
            

class Chatterbox:
    def __init__(self, svr: str, port: int) -> None:
        self.svr = svr
        self.port = port
        self.sok = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.pub = Delivery()
        self.pub.subscribe('WORKER', self.obs)
        self.still = True

    def connect(self) -> None:
        try:
            self.sok.connect((self.svr, self.port))
            msg = ['OUTCOME', f'Connessione con {self.svr} stabilita']
            self.pub.send_message('OBSERVER', msg)
            # lancia un thread per l'ascolto
            t = Thread(target=self.listen)
            t.daemon = True
            t.start()
        except OSError as e:
            msg = ['ERROR', repr(e)]
            self.pub.send_message('OBSERVER', msg)

    def listen(self) -> None:
        while self.still:
            data = self.sok.recv(2048)
            # qui andrebbe una queue per collezionare i dati giunti
            # ma complicherebbe le cose, mi limito a comunicare
            msg = ['RECEIVED', data.decode()]
            self.pub.send_message('LISTENERS', msg)
        self.sok.close()
        msg = ['OUTCOME', f'Connessione con {self.svr} chiusa']
        self.pub.send_message('OBSERVER', msg)
        self.pub.unsubscribe('WORKER', self.obs)

    def transmit(self, message: bytes) -> None:
        try:
            self.sok.sendall(message)
        except OSError as e:
            msg = ['ERROR', repr(e)]
            self.pub.send_message('OBSERVER', msg)
    
    def close(self) -> None:
        self.still = False

    def obs(self, message):
        op = message[0]
        if op == 'CLOSING' and message[0] == 'CLOSING':
            self.still = False
        
        
